#include <cstring>
#include <iostream>
#include <fstream>
#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctime>

#include "sbcustomcertstorage.h"
#include "sbmisc.h"
#include "sboffice.h"
#include "sbofficesecurity.h"
#include "sbutils.h"
#include "sbstrutils.h"
#include "sbx509.h"
#include "sbxmlades.h"
#include "sbhttptspclient.h"
#include "sbhttpsclient.h"

using namespace SecureBlackbox;

#define OP_VERIFY	0
#define OP_SIGN		1
#define OP_ENCRYPT	2
#define OP_DECRYPT	3

int LoadCertificate(TElX509Certificate &cert, std::string filename, std::string password)
{
	std::ifstream fs;
	fs.open(filename.c_str(), std::ios::binary | std::ios::in);
	if (fs.is_open())
	{
		TElCallbackStream cs(fs);
		int k = cert.LoadFromStreamAuto(cs, password, 0);
		fs.close();
		if (k != 0)
			std::cout << std::hex << "Failed to load certificate: 0x" << k << std::endl;

		return k;
	}
	else
	{
		std::cout << "Unable to open file: " << filename << std::endl;
		return 1;
	}
}

bool CopyFile(std::string fromFilename, std::string toFilename)
{
	std::ifstream src;
	std::ofstream dst;
	src.open(fromFilename.c_str(), std::ios::binary | std::ios::in);
	if (!src.is_open())
	{
		std::cout << "Unable to open file: " << fromFilename << std::endl;
		return false;
	}

	dst.open(toFilename.c_str(), std::ios::binary | std::ios::out);
	if (!dst.is_open())
	{
		std::cout << "Unable to open file: " << toFilename << std::endl;
		return false;
	}

	dst << src.rdbuf();

	dst.close();
	src.close();
	return true;
}

void SignDocument(TElOfficeDocument &Document, TElX509Certificate &Cert, bool signDocumentPartOnly, bool signCoreProperties, bool useCryptoAPIHandler, std::string &tsaURL)
{
	try
	{
		if (!Document.get_Signable())
		{
			std::cout << "Cannot sign document" << std::endl;
			return;
		}

		TElHTTPTSPClient tspClient(NULL);
		TElHTTPSClient httpClient(NULL);
		if (tsaURL != "")
		{
			tspClient.set_HTTPClient(httpClient);
			tspClient.set_URL(tsaURL);
			tspClient.set_HashAlgorithm(SB_ALGORITHM_DGST_SHA1);
			std::cout << "Timestamp server: " << tsaURL << std::endl;
		}

		if (Document.get_OpenXMLDocument() != NULL)
		{
			TElOfficeOpenXMLSignatureHandler OpenXMLSigHandler(NULL);
			Document.AddSignature(OpenXMLSigHandler, true); // add a signature handler, it will be freed automatically on document Close
			if (!signDocumentPartOnly)
				OpenXMLSigHandler.AddDocument();
			else
				OpenXMLSigHandler.AddPart(Document.get_OpenXMLDocument()->get_OfficeDocumentPart());

			if (signCoreProperties)
				OpenXMLSigHandler.AddCoreProperties();

			OpenXMLSigHandler.get_SignatureInfoV1()->set_SignatureText(L"Signed by SecureBlackbox");

			if (tsaURL != "")
			{
				OpenXMLSigHandler.GenerateXAdES(XAdES_T, Cert, tspClient);
			}

			OpenXMLSigHandler.Sign(Cert);

			OpenXMLSigHandler.updateHandle(NULL); // detach wrapper class 
		}
		else if (Document.get_OpenXPSDocument() != NULL)
		{
			TElOfficeOpenXPSSignatureHandler OpenXPSSigHandler(NULL);
			Document.AddSignature(OpenXPSSigHandler, true);

			OpenXPSSigHandler.AddDocument();
			if (signCoreProperties)
				OpenXPSSigHandler.AddCoreProperties();

			if (tsaURL != "")
			{
				OpenXPSSigHandler.GenerateXAdES(XAdES_T, Cert, tspClient);
			}

			OpenXPSSigHandler.Sign(Cert);

			OpenXPSSigHandler.updateHandle(NULL);
		}
		else if (Document.get_BinaryDocument() != NULL)
		{
			if (useCryptoAPIHandler)
			{
				TElOfficeBinaryCryptoAPISignatureHandler BinCryptoAPISigHandler(NULL);
				Document.AddSignature(BinCryptoAPISigHandler, true);
				std::time_t t;
				std::time(&t);
				t += 24 * 60 * 60; // add 1 day
				BinCryptoAPISigHandler.set_ExpireTime(t);

				if (tsaURL != "")
					std::cout << "Timestamp not applicable" << std::endl;

				BinCryptoAPISigHandler.Sign(Cert);

				BinCryptoAPISigHandler.updateHandle(NULL);
			}
			else
			{
				TElOfficeBinaryXMLSignatureHandler BinXMLSigHandler(NULL);
				Document.AddSignature(BinXMLSigHandler, true);

				BinXMLSigHandler.get_SignatureInfoV1()->set_SignatureText(L"Signed by SecureBlackbox");

				if (tsaURL != "")
				{
					BinXMLSigHandler.GenerateXAdES(XAdES_T, Cert, tspClient);
				}

				BinXMLSigHandler.Sign(Cert);

				BinXMLSigHandler.updateHandle(NULL);
			}
		}
		else if (Document.get_OpenDocument() != NULL)
		{
			TElOpenOfficeSignatureHandler ODFSigHandler (NULL);
			Document.AddSignature(ODFSigHandler, true);

			ODFSigHandler.AddDocument();

			if (tsaURL != "")
			{
				ODFSigHandler.GenerateXAdES(XAdES_T, Cert, tspClient);
			}

			ODFSigHandler.Sign(Cert);

			ODFSigHandler.updateHandle(NULL);
		}
		else
		{
			std::cout << "Failed to sign" << std::endl;
			return;
		}

		std::cout << "Signing succeeded" << std::endl;
	}
	catch (SBException E)
	{
		std::cout << E.what() << std::endl;
		std::cout << "Stack trace: " << E.getErrorStackTrace().c_str() << std::endl;
	}
}

void VerifyDocument(TElOfficeDocument &Document)
{
	try
	{
		while (true)
		{
			std::cout << Document.get_SignatureHandlerCount() << " signatures is availble" << std::endl;
			std::cout << "Please select a signature to validate (1 - first signature, 0 - to exit): " << std::endl;
			std::string input;
			std::getline(std::cin, input);
			int k = atoi(input.c_str()) - 1;
			if (k == -1)
				return;

			if ((k < 0) || (k >= Document.get_SignatureHandlerCount()))
			{
				std::cout << "Invalid signature index (exit)" << std::endl;
				return;
			}

			TElX509Certificate SignerCert(NULL);
			TElOfficeCustomSignatureHandler *pHandler = Document.get_SignatureHandlers(k);
			if (pHandler->ClassNameIs_Inst("TElOfficeBinaryXMLSignatureHandler"))
			{
				TElOfficeBinaryXMLSignatureHandler BinXMLSigHandler(pHandler->getHandle(), false); // casting a type
				TSBOfficeBinarySignatureValidationStatus BinValidationStatus = BinXMLSigHandler.Validate();
				if (bsvsValid == BinValidationStatus)
					std::cout << "Signature is VALID" << std::endl;
				else if (bsvsValidButNotEntries == BinValidationStatus)
				{
					std::cout << "Signature integrity is ok, but some of entries are INVALID" << std::endl;
					for (int i = 0; i < BinXMLSigHandler.get_SignedEntryCount(); i++)
					{
						TElOfficeBinarySignedEntry *pEntry = BinXMLSigHandler.get_SignedEntries(i);
						std::wstring s;
						pEntry->get_Path(s);
						std::wcout << L"\t" << s;
						TSBOfficeBinaryEntryValidationStatus Status = pEntry->get_ValidationStatus();
						if (bevsValid == Status)
							std::cout << "Valid";
						else if (bevsInvalidDigest == Status)
							std::cout << "Invalid Digest";
						else if (bevsInvalidSignature == Status)
							std::cout << "Invalid Signature";
						else if (bevsEntryIsMissing == Status)
							std::cout << "Entry Is Missing";
						else if (bevsEntryNotSigned == Status)
							std::cout << "Entry Not Signed";
						else
							std::cout << "Unknown";

						std::cout << std::endl;
					}
				}
				else
					std::cout << "Signature is INVALID" << std::endl;

				time_t timer = (time_t)BinXMLSigHandler.get_SignatureTime()->get_ValueUTC();
				char* dt = ctime(&timer);
				std::cout << "Signature time: " << dt << std::endl;
				if (BinXMLSigHandler.get_SignerCertificate() != NULL)
					BinXMLSigHandler.get_SignerCertificate()->Clone(SignerCert, false);
			}
			else if (pHandler->ClassNameIs_Inst("TElOfficeBinaryCryptoAPISignatureHandler"))
			{
				TElOfficeBinaryCryptoAPISignatureHandler BinCryptoAPISigHandler(pHandler->getHandle(), false);
				TSBOfficeBinarySignatureValidationStatus BinValidationStatus = BinCryptoAPISigHandler.Validate();
				if (bsvsValid == BinValidationStatus)
					std::cout << "Signature is VALID" << std::endl;
				else
					std::cout << "Signature is INVALID" << std::endl;

				time_t timer = (time_t)BinCryptoAPISigHandler.get_SignTime();
				char* dt = ctime(&timer);
				std::cout << "Sign time: " << dt << std::endl;

				timer = (time_t)BinCryptoAPISigHandler.get_ExpireTime();
				dt = ctime(&timer);
				std::cout << "Expire time: " << dt << std::endl;
				if (BinCryptoAPISigHandler.get_Certificate() != NULL)
					BinCryptoAPISigHandler.get_Certificate()->Clone(SignerCert, false);
			}
			else if (pHandler->ClassNameIs_Inst("TElOfficeOpenXMLBaseSignatureHandler") ||
				pHandler->ClassNameIs_Inst("TElOfficeOpenXMLSignatureHandler") ||
				pHandler->ClassNameIs_Inst("TElOfficeOpenXPSSignatureHandler"))
			{
				TElOfficeOpenXMLBaseSignatureHandler OpenBaseSigHandler(pHandler->getHandle(), false);
				TSBOfficeOpenXMLSignatureValidationStatus ValidationStatus = OpenBaseSigHandler.Validate();
				if (svsValid == ValidationStatus)
					std::cout << "Signature is VALID" << std::endl;
				else if (svsValidButNotParts == ValidationStatus)
				{
					std::cout << "Signature integrity is ok, but some of parts are INVALID" << std::endl;
					for (int i = 0; i < OpenBaseSigHandler.get_SignedPartCount(); i++)
					{
						TElOfficeOpenXMLSignedPart *pEntry = OpenBaseSigHandler.get_SignedParts(i);
						std::string s;
						pEntry->get_URI(s);
						std::cout << "\t" << s;
						TSBOfficeOpenXMLPartValidationStatus Status = pEntry->get_ValidationStatus();
						if (pvsValid == Status)
							std::cout << "Valid";
						else if (pvsInvalidDigest == Status)
							std::cout << "Invalid Digest";
						else if (pvsInvalidSignature == Status)
							std::cout << "Invalid Signature";
						else if (pvsPartIsMissing == Status)
							std::cout << "Part Is Missing";
						else if (pvsInvalidContentType == Status)
							std::cout << "Invalid ContentType";
						else
							std::cout << "Unknown";

						std::cout << std::endl;
					}
				}
				else
					std::cout << "Signature is INVALID" << std::endl;

				time_t timer = (time_t)OpenBaseSigHandler.get_SignatureTime()->get_ValueUTC();
				char* dt = ctime(&timer);
				std::cout << "Signature time: " << dt << std::endl;
				if (OpenBaseSigHandler.get_SignerCertificate() != NULL)
					OpenBaseSigHandler.get_SignerCertificate()->Clone(SignerCert, false);
			}
			else if (pHandler->ClassNameIs_Inst("TElOpenOfficeSignatureHandler"))
			{
				TElOpenOfficeSignatureHandler ODFSigHandler(pHandler->getHandle(), false);
				TSBOpenOfficeSignatureValidationStatus ODFValidationStatus = ODFSigHandler.Validate();
				if (osvsValid == ODFValidationStatus)
					std::cout << "Signature is VALID" << std::endl;
				else if (osvsValidButNotEntries == ODFValidationStatus)
				{
					std::cout << "Signature integrity is ok, but some of entries are INVALID" << std::endl;
					for (int i = 0; i < ODFSigHandler.get_SignedEntryCount(); i++)
					{
						TElOpenOfficeSignedEntry *pEntry = ODFSigHandler.get_SignedEntries(i);
						std::string s;
						pEntry->get_Path(s);
						std::cout << "\t" << s;
						TSBOpenOfficeEntryValidationStatus Status = pEntry->get_ValidationStatus();
						if (oevsValid == Status)
							std::cout << "Valid";
						else if (oevsInvalidDigest == Status)
							std::cout << "Invalid Digest";
						else if (oevsInvalidSignature == Status)
							std::cout << "Invalid Signature";
						else if (oevsEntryIsMissing == Status)
							std::cout << "Entry Is Missing";
						else
							std::cout << "Unknown";

						std::cout << std::endl;
					}
				}
				else
					std::cout << "Signature is INVALID" << std::endl;

				time_t timer = (time_t)ODFSigHandler.get_SignatureTime()->get_ValueUTC();
				char* dt = ctime(&timer);
				std::cout << "Signature time: " << dt << std::endl;
				if (ODFSigHandler.get_SignerCertificate() != NULL)
					ODFSigHandler.get_SignerCertificate()->Clone(SignerCert, false);
			}
			else if (pHandler->ClassNameIs_Inst("TElOfficeBinaryInvalidSignatureHandler"))
			{
				TElOfficeBinaryInvalidSignatureHandler InvalidHandler(pHandler->getHandle(), false);
				std::string msg;
				InvalidHandler.get_ErrorMessage(msg);
				std::cout << "Error message: " << msg << std::endl;
			}
			else if (pHandler->ClassNameIs_Inst("TElOfficeOpenXMLInvalidSignatureHandler"))
			{
				TElOfficeOpenXMLInvalidSignatureHandler InvalidHandler(pHandler->getHandle(), false);
				std::string msg;
				InvalidHandler.get_ErrorMessage(msg);
				std::cout << "Error message: " << msg << std::endl;
			}
			else if (pHandler->ClassNameIs_Inst("TElOpenOfficeInvalidSignatureHandler"))
			{
				TElOpenOfficeInvalidSignatureHandler InvalidHandler(pHandler->getHandle(), false);
				std::string msg;
				InvalidHandler.get_ErrorMessage(msg);
				std::cout << "Error message: " << msg << std::endl;
			}

			std::cout << "Signer certificate: " << std::endl;
			TName name;
			SignerCert.get_SubjectName(name);
			std::cout << "\tSubject: ";
			if (name.CommonName != NULL)
				std::cout << (char *)name.CommonName;
			else
				std::cout << (char *)name.Organization;

			std::cout << std::endl;
			SignerCert.get_IssuerName(name);
			std::cout << "\tIssuer: ";
			if (name.CommonName != NULL)
				std::cout << (char *)name.CommonName;
			else
				std::cout << (char *)name.Organization;

			std::cout << std::endl;
			std::vector<uint8_t> SerialNumber;
			SignerCert.get_SerialNumber(SerialNumber);
			std::string s;
			BinaryToString(SerialNumber, s);
			BeautifyBinaryString(s, ':', s);
			std::cout << "\tSerial number: " << s << std::endl;

			std::cout << std::endl;
		}
	}
	catch (SBException E)
	{
		std::cout << E.what() << std::endl;
		std::cout << "Stack trace: " << E.getErrorStackTrace().c_str() << std::endl;
	}
}

void EncryptDocument(TElOfficeDocument &Document, std::string &OutputFilename, std::string &Password, bool useCryptoAPIHandler, bool useAgileHandler, int32_t EncryptionAlgorithm, int32_t HashAlgorithm)
{
	try
	{
		std::wstring wPassword; // some handlers require unicode password and some ANSI
		std::vector<uint8_t> Buf;
		BytesOfString(Password, Buf);
		UTF8ToWideStr(Buf, wPassword);		

		TElOfficeCustomEncryptionHandler *Handler = NULL;
		if (dfBinary == Document.get_DocumentFormat())
		{
			if (useCryptoAPIHandler)
			{
				TElOfficeBinaryRC4CryptoAPIEncryptionHandler *RC4CryptoAPIEncryptionHandler = new TElOfficeBinaryRC4CryptoAPIEncryptionHandler(NULL);
				Handler = RC4CryptoAPIEncryptionHandler;
				RC4CryptoAPIEncryptionHandler->set_KeyLength(128);
				RC4CryptoAPIEncryptionHandler->set_HardenedKeyGeneration(true);
				RC4CryptoAPIEncryptionHandler->set_EncryptDocumentProperties(true);
				RC4CryptoAPIEncryptionHandler->set_Password(wPassword);
			}
			else
			{
				TElOfficeBinaryRC4EncryptionHandler *RC4EncryptionHandler = new TElOfficeBinaryRC4EncryptionHandler(NULL);
				Handler = RC4EncryptionHandler;
				RC4EncryptionHandler->set_Password(wPassword);
			}
		}		
		else if (dfOpenXML == Document.get_DocumentFormat())
		{
			if (useAgileHandler)
			{
				TElOfficeOpenXMLAgileEncryptionHandler *AgileEncryptionHandler = new TElOfficeOpenXMLAgileEncryptionHandler(NULL);
				Handler = AgileEncryptionHandler;
				AgileEncryptionHandler->set_CipherAlgorithm(EncryptionAlgorithm);
				AgileEncryptionHandler->set_HashAlgorithm(HashAlgorithm);
				AgileEncryptionHandler->set_SaltSize(16);
				int k = AgileEncryptionHandler->AddPasswordKeyEncryptor(wPassword);
				TElOfficeOpenXMLPasswordKeyEncryptor KeyEncryptor(AgileEncryptionHandler->get_KeyEncryptors(k)->getHandle(), false);
				KeyEncryptor.set_CipherAlgorithm(SB_ALGORITHM_CNT_AES128);
				KeyEncryptor.set_HashAlgorithm(SB_ALGORITHM_DGST_SHA1);
				KeyEncryptor.set_SaltSize(16);
				KeyEncryptor.set_SpinCount(100000);
			}
			else
			{
				TElOfficeOpenXMLStandardEncryptionHandler *StandardEncryptionHandler = new TElOfficeOpenXMLStandardEncryptionHandler(NULL);
				Handler = StandardEncryptionHandler;
				StandardEncryptionHandler->set_EncryptionAlgorithm(EncryptionAlgorithm);
				StandardEncryptionHandler->set_Password(wPassword);
			}
		}
		else if (dfOpenDocument == Document.get_DocumentFormat())
		{
			TElOpenOfficeEncryptionHandler *ODFEncryptionHandler = new TElOpenOfficeEncryptionHandler(NULL);
			Handler = ODFEncryptionHandler;
			ODFEncryptionHandler->set_EncryptionAlgorithm(EncryptionAlgorithm);
			ODFEncryptionHandler->set_Password(Password);
		}
		else
		{
			std::cout << "Document cannot be encrypted" << std::endl;
			return;
		}

		if (Handler != NULL)
		{
			TFileStream fs(OutputFilename, filemodeCreate);

			Document.set_EncryptionHandler(Handler);
			Document.EncryptTo(fs);
			std::cout << "The document encrypted OK" << std::endl;

			delete Handler;
		}
	}
	catch (SBException E)
	{
		std::cout << E.what() << std::endl;
		std::cout << "Stack trace: " << E.getErrorStackTrace().c_str() << std::endl;
	}
}

void DecryptDocument(TElOfficeDocument &Document, std::string &OutputFilename, std::string &Password)
{
	try
	{
		if (!Document.get_IsEncrypted())
		{
			std::cout << "The document is not encrypted" << std::endl;
			return;
		}

		std::wstring wPassword; // some handlers require unicode password and some ANSI
		std::vector<uint8_t> Buf;
		BytesOfString(Password, Buf);
		UTF8ToWideStr(Buf, wPassword);

		if (Document.get_EncryptionHandler() == NULL)
		{
			std::cout << "Unknown encryption handler" << std::endl;
			return;			
		}

		TElOfficeCustomEncryptionHandler *Handler = Document.get_EncryptionHandler();
		if (Handler->ClassNameIs_Inst("TElOfficeOpenXMLStandardEncryptionHandler"))
		{
			TElOfficeOpenXMLStandardEncryptionHandler StandardEncryptionHandler(Handler->getHandle(), false);
			std::string s;
			GetAlgorithmNameByAlgorithm(StandardEncryptionHandler.get_EncryptionAlgorithm(), s);
			std::cout << " Encryption algorithm: " << s << std::endl;

			StandardEncryptionHandler.set_Password(wPassword);
			if (!StandardEncryptionHandler.IsPasswordValid())
			{
				std::cout << "Invalid password" << std::endl;
				return;
			}
		}
		else if (Handler->ClassNameIs_Inst("TElOfficeOpenXMLAgileEncryptionHandler"))
		{
			TElOfficeOpenXMLAgileEncryptionHandler AgileEncryptionHandler(Handler->getHandle(), false);
			std::string s;
			std::cout << " Key Encryptors: " << AgileEncryptionHandler.get_KeyEncryptorCount() << std::endl;
			GetAlgorithmNameByAlgorithm(AgileEncryptionHandler.get_CipherAlgorithm(), s);
			std::cout << " Cipher algorithm: " << s << std::endl;
			GetAlgorithmNameByAlgorithm(AgileEncryptionHandler.get_HashAlgorithm(), s);
			std::cout << " Hash algorithm: " << s << std::endl;

			bool found = false;
			for (int i = 0; i < AgileEncryptionHandler.get_KeyEncryptorCount(); i++)
			{
				if (AgileEncryptionHandler.get_KeyEncryptors(i)->ClassNameIs_Inst("TElOfficeOpenXMLPasswordKeyEncryptor"))
				{
					TElOfficeOpenXMLPasswordKeyEncryptor KeyEncryptor(AgileEncryptionHandler.get_KeyEncryptors(i)->getHandle(), false);
					KeyEncryptor.set_Password(wPassword);
					if (KeyEncryptor.IsPasswordValid())
					{
						found = true;
						break;
					}
				}
			}

			if (!found)
			{
				std::cout << "Invalid password" << std::endl;
				return;
			}
		}
		else if (Handler->ClassNameIs_Inst("TElOfficeBinaryRC4EncryptionHandler"))
		{
			TElOfficeBinaryRC4EncryptionHandler RC4EncryptionHandler(Handler->getHandle(), false);
			RC4EncryptionHandler.set_Password(wPassword);
			if (!RC4EncryptionHandler.IsPasswordValid())
			{
				std::cout << "Invalid password" << std::endl;
				return;
			}
		}
		else if (Handler->ClassNameIs_Inst("TElOfficeBinaryRC4CryptoAPIEncryptionHandler"))
		{
			TElOfficeBinaryRC4CryptoAPIEncryptionHandler RC4CryptoAPIEncryptionHandler(Handler->getHandle(), false);
			std::cout << " Key length: " << RC4CryptoAPIEncryptionHandler.get_KeyLength() << std::endl;
			if (RC4CryptoAPIEncryptionHandler.get_EncryptDocumentProperties())
				std::cout << " Document properties are encrypted" << std::endl;
			else
				std::cout << " Document properties are NOT encrypted" << std::endl;

			RC4CryptoAPIEncryptionHandler.set_Password(wPassword);
			if (!RC4CryptoAPIEncryptionHandler.IsPasswordValid())
			{
				std::cout << "Invalid password" << std::endl;
				return;
			}
		}
		else if (Handler->ClassNameIs_Inst("TElOpenOfficeEncryptionHandler"))
		{
			TElOpenOfficeEncryptionHandler ODFEncryptionHandler(Handler->getHandle(), false);
			ODFEncryptionHandler.set_Password(Password);
			if (!ODFEncryptionHandler.IsPasswordValid())
			{
				std::cout << "Invalid password" << std::endl;
				return;
			}
		}
		else
		{
			std::cout << "Unknown encryption handler" << std::endl;
			return;
		}

		TFileStream fs(OutputFilename, filemodeCreate);
		Document.DecryptTo(fs);
		std::cout << "The document decrypted OK" << std::endl;
	}
	catch (SBException E)
	{
		std::cout << E.what() << std::endl;
		std::cout << "Stack trace: " << E.getErrorStackTrace().c_str() << std::endl;
	}
}

int main(int argc, char **argv) {

	try
	{
		#error Please pick the evaluation license key from <SecureBlackbox>\LicenseKey.txt file and place it here. If the evaluation key expires, you can request an extension using the form on https://www.eldos.com/sbb/keyreq/
SetLicenseKey("<put the key here>");
	}
	catch (SBException E)
	{
		std::cout << "Failed to set license key:" << std::endl;
		std::cout << E.what() << std::endl;
		return 0;
	}

	if (argc < 2)
	{
		std::cout << "Usage:" << std::endl;
		std::cout << "ConsoleOffice [operation] [signature/encryption options] [key options] input_file [-out output_file]" << std::endl <<
			"Operation: " << std::endl <<
			" -sign : sign Office document" << std::endl <<
			" -verify : verify Office document" << std::endl <<
			" -encrypt : encrypt Office document" << std::endl <<
			" -decrypt : decrypt Office document" << std::endl <<
			"Signature options: " << std::endl <<
			" -tsa <url> : timestamp server;" << std::endl <<
			" -sign-doc-part : sign document part only (for OOXML documents)" << std::endl <<
			" -sign-core-prop : sign core properties (for OOXML and XPS documents)" << std::endl <<
			" -use-crypto-api : use CryptoAPI signature handler (for binary doucments)" << std::endl <<
			"Key options:" << std::endl <<
			" -cert filename : certificate filename" << std::endl <<
			" -certpass password : certificate password" << std::endl <<
			"Encryption options: " << std::endl <<
			" -pass password : encryption/decryption password" << std::endl <<
			" -use-crypto-api : use CryptoAPI encryption handler (for binary doucments)" << std::endl <<
			" -use-agile : use Agile encryption handler (for OOXML doucments)" << std::endl <<
			" -aes128|aes192|aes256|des|3des|rc2|rc4 : encryption algorithm (for OOXML and ODF documents)" << std::endl <<
			" -sha1|sha256|sha384|sha512|md5|ripemd160 : hash algorithm (for OOXML documents with agile handler)" << std::endl <<
			std::endl;

		return 0;
	}

	// command line parameters
	int paramOperation = OP_VERIFY;
	std::string paramInputFile;
	std::string paramOutputFile;
	std::string paramCertFile;
	std::string paramCertPassword;
	std::string paramPassword;
	std::string paramTSAURL;
	bool paramSignDocPart = false;
	bool paramSignCoreProp = false;
	bool paramUseCryptoAPI = false;
	bool paramUseAgile = false;
	int32_t paramEncryptionAlg = SB_ALGORITHM_CNT_AES128;
	int32_t paramHashAlg = SB_ALGORITHM_DGST_SHA1;

	int i = 1;
	while (i < argc - 1)
	{
		char *arg = argv[i];
		if (!strcmp(arg, "-sign"))
			paramOperation = OP_SIGN;
		else if (!strcmp(arg, "-verify"))
			paramOperation = OP_VERIFY;
		else if (!strcmp(arg, "-encrypt"))
			paramOperation = OP_ENCRYPT;
		else if (!strcmp(arg, "-decrypt"))
			paramOperation = OP_DECRYPT;
		else if (!strcmp(arg, "-cert"))
			paramCertFile.assign(argv[++i]);
		else if (!strcmp(arg, "-certpass"))
			paramCertPassword.assign(argv[++i]);
		else if (!strcmp(arg, "-pass"))
			paramPassword.assign(argv[++i]);
		else if (!strcmp(arg, "-tsa"))
			paramTSAURL.assign(argv[++i]);
		else if (!strcmp(arg, "-sign-doc-part"))
			paramSignDocPart = true;
		else if (!strcmp(arg, "-sign-core-prop"))
			paramSignCoreProp = true;
		else if (!strcmp(arg, "-use-crypto-api"))
			paramUseCryptoAPI = true;	
		else if (!strcmp(arg, "-use-agile"))
			paramUseAgile = true;
		else if (!strcmp(arg, "-aes128"))
			paramEncryptionAlg = SB_ALGORITHM_CNT_AES128;
		else if (!strcmp(arg, "-aes192"))
			paramEncryptionAlg = SB_ALGORITHM_CNT_AES192;
		else if (!strcmp(arg, "-aes256"))
			paramEncryptionAlg = SB_ALGORITHM_CNT_AES256;
		else if (!strcmp(arg, "-des"))
			paramEncryptionAlg = SB_ALGORITHM_CNT_DES;
		else if (!strcmp(arg, "-3des"))
			paramEncryptionAlg = SB_ALGORITHM_CNT_3DES;
		else if (!strcmp(arg, "-rc2"))
			paramEncryptionAlg = SB_ALGORITHM_CNT_RC2;
		else if (!strcmp(arg, "-rc4"))
			paramEncryptionAlg = SB_ALGORITHM_CNT_RC4;
		else if (!strcmp(arg, "-sha1"))
			paramHashAlg = SB_ALGORITHM_DGST_SHA1;
		else if (!strcmp(arg, "-sha256"))
			paramHashAlg = SB_ALGORITHM_DGST_SHA256;
		else if (!strcmp(arg, "-sha384"))
			paramHashAlg = SB_ALGORITHM_DGST_SHA384;
		else if (!strcmp(arg, "-sha512"))
			paramHashAlg = SB_ALGORITHM_DGST_SHA512;
		else if (!strcmp(arg, "-md5"))
			paramHashAlg = SB_ALGORITHM_DGST_MD5;
		else if (!strcmp(arg, "-ripemd160"))
			paramHashAlg = SB_ALGORITHM_DGST_RIPEMD160;
		else if (!strcmp(arg, "-out"))
			paramOutputFile.assign(argv[++i]);
		else
			paramInputFile.assign(argv[i]);

		i++;
	}

	if (i < argc)
		paramInputFile.assign(argv[i]);

	if (paramInputFile.empty())
	{
		std::cout << "Input filename is not specified" << std::endl;
		return 0;
	}

	if (paramOperation != OP_VERIFY)
	{
		if (paramOutputFile.empty())
		{
			std::cout << "Output filename is not specified" << std::endl;
			return 0;
		}
	}

	if (OP_SIGN == paramOperation)
	{
		if (paramCertFile.empty())
		{
			std::cout << "Certificate filename is not specified" << std::endl;
			return 0;
		}
	}

	TElOfficeDocument Document(NULL);

	try
	{
		std::cout << "Input Filename: " << paramInputFile << std::endl;
		if (OP_SIGN == paramOperation)
		{
			CopyFile(paramInputFile, paramOutputFile);
			std::cout << "Output Filename: " << paramOutputFile << std::endl;
			Document.Open(paramOutputFile, false);
		}
		else
		{
			Document.Open(paramInputFile, true);
		}
	}
	catch (SBException E)
	{
		std::cout << "Failed to load Office document: " << E.what() << std::endl;
		std::cout << "Stack trace: " << E.getErrorStackTrace().c_str() << std::endl;
		return 0;
	}

	try
	{
		std::string s;
		DocumentFormatToString(Document.get_DocumentFormat(), s);
		std::cout << "Document format: " << s << std::endl;
		if (Document.get_BinaryDocument() != NULL)
		{
			BinaryDocumentTypeToString(Document.get_BinaryDocument()->get_DocumentType(), s);
			std::cout << "Document type: " << s << std::endl;
		}
		else if (Document.get_OpenXMLDocument() != NULL)
		{
			OpenXMLDocumentTypeToString(Document.get_OpenXMLDocument()->get_DocumentType(), s);
			std::cout << "Document type: " << s << std::endl;
		}

		if (Document.get_IsSigned())
		{
			std::cout << "Document is SIGNED" << std::endl;
		}

		if (Document.get_IsEncrypted())
		{
			if (Document.get_EncryptionHandler() != NULL)
			{
				Document.get_EncryptionHandler()->GetName_Inst(s);
				std::cout << "The document is ENCRYPTED using '" << s << "' handler" << std::endl;
			}
			else
				std::cout << "The document is ENCRYPTED using UNKNOWN handler" << std::endl;
		}

		if (!Document.get_IsEncrypted() && !Document.get_IsSigned())
			std::cout << "Document is neither encrypted nor signed" << std::endl;

		switch (paramOperation)
		{
		case OP_SIGN:
		{
			TElX509Certificate Cert(NULL);
			if (LoadCertificate(Cert, paramCertFile, paramCertPassword))
			{
				Document.Close();
				return 0;
			}
				
			SignDocument(Document, Cert, paramSignDocPart, paramSignCoreProp, paramUseCryptoAPI, paramTSAURL);
			break;
		}
		case OP_ENCRYPT:
		{
			EncryptDocument(Document, paramOutputFile, paramPassword, paramUseCryptoAPI, paramUseAgile, paramEncryptionAlg, paramHashAlg);
			break;
		}
		case OP_DECRYPT:
		{
			DecryptDocument(Document, paramOutputFile, paramPassword);
			break;
		}
		default:
			if (Document.get_IsSigned())
				VerifyDocument(Document);

			break;
		}

		// close a document
		Document.Close();
	}
	catch (SBException E)
	{
		std::cout << E.what() << std::endl;
		std::cout << "Stack trace: " << E.getErrorStackTrace().c_str() << std::endl;
		return 0;
	}

	return 0;
}
