#include <cstring>
#include <iostream>
#include <fstream>
#include <stdio.h>
#include <stdlib.h>
#include <algorithm>
#include <ctime>

#include "sbpgp.h"
#include "sbpgpkeys.h"
#include "sbmisc.h"

using namespace SecureBlackbox;

std::string outputPath;

std::string UnQuote(std::string s)
{
    if (s.length() < 1)
        return s;

	s.erase(remove( s.begin(), s.end(), '\"' ), s.end());

    return s;
}

std::string filename(std::string const& path)
{
	std::string result;
	size_t pos = path.find_last_of("\\");
	if(pos != std::string::npos)
		result.assign(path.begin() + pos + 1, path.end());
	else
		result = path;

	return result;
}

std::string ReadLine()
{
	std::string result;
	std::getline(std::cin, result);
	return result;
}

void SB_CALLBACK PGP_KeyPassphraseEvent(void * /*_ObjectData*/, TObjectHandle /*Sender*/, TElPGPCustomSecretKeyHandle Key, char * pcPassphrase, int32_t * szPassphrase, int8_t * Cancel)
{
	if (Key != NULL)
	{
		std::cout << "Passphrase is needed for secret key:" << std::endl;
	}
	else
	{
		std::cout << "Passphrase is needed to decrypt message" << std::endl;
	}

	std::string password = ReadLine();
	try
	{
		SBCheckError(SBSetEventFirstReturnStringA(password.data(), password.length()));
    }
    catch (SBException ex)
    {
		std::cout << ex.what() << std::endl;
    }

	*Cancel = false;
}

void EncryptAndSign(TStringList &inputFiles, const std::string &strOutputFilename, TElPGPKeyring &Keyring, const std::string &password)
{
	TElPGPWriter pgpWriter (NULL);
	pgpWriter.set_Armor(true);
	
	TStringList* armorHeaders = pgpWriter.get_ArmorHeaders();
	armorHeaders->Clear();
	armorHeaders->Add("Version: EldoS OpenPGPBlackbox");
	
	pgpWriter.set_ArmorBoundary("PGP MESSAGE");
	
	if (Keyring.get_PublicCount() > 0)
		pgpWriter.set_EncryptingKeys(Keyring);

	if (Keyring.get_SecretCount() > 0)
		pgpWriter.set_SigningKeys(Keyring);

	if (password != "")
		pgpWriter.get_Passphrases()->Add(password);

	if (password == "")
		pgpWriter.set_EncryptionType(etPublicKey);
	else if (Keyring.get_PublicCount() == 0)
		pgpWriter.set_EncryptionType(etPassphrase);
	else 
		pgpWriter.set_EncryptionType(etBoth);

	pgpWriter.set_OnKeyPassphrase(&PGP_KeyPassphraseEvent, NULL);

	std::time_t t = std::time(0);
	pgpWriter.set_Timestamp((int64_t)t);

	if (inputFiles.get_Count() == 1) // process single file
    {
		std::string strInputFilename;
		inputFiles.get_Strings(0, strInputFilename);

        pgpWriter.set_Filename(filename(strInputFilename));
		TFileStream streamInput (strInputFilename, filemodeOpenReadWrite);
        try
        {
			TFileStream streamOutput (strOutputFilename, filemodeCreate);

			pgpWriter.EncryptAndSign(streamInput, streamOutput, 0);
            std::cout << "The file was encrypted successfully" << std::endl;
        }
        catch (SBException ex)
        {
			std::cout << ex.what() << std::endl;
        }
    }
    else
    {
        // create list of input file streams
		TList inFileStreams;
        try
        {
			// create filestream for writing encrypted file
			TFileStream streamOutput (strOutputFilename, filemodeCreate);

			for (int i = 0; i < inputFiles.get_Count(); i++)
			{
				// add input streams to list
				std::string strInputFilename;
				inputFiles.get_Strings(i, strInputFilename);
				TFileStream streamInput (strInputFilename, filemodeOpenReadWrite);
				inFileStreams.Add(streamInput.getHandle());
				streamInput.updateHandle(NULL);
				// create input file names list
				inputFiles.set_Strings(i, filename(strInputFilename));
			}

			std::string outputFilename = filename(strOutputFilename);
			pgpWriter.EncryptAndSignFiles(inputFiles, inFileStreams, outputFilename, streamOutput);
            std::cout << "The files were encrypted successfully" << std::endl;

			for (int i = 0; i < inFileStreams.get_Count(); i++)
			{
				// clear filestream handles
				TObjectHandle handle = (TObjectHandle)inFileStreams.get_Items(i);
				TObject_Free(handle);
			}
        }
        catch (SBException ex)
        {
            std::cout << ex.what() << std::endl;
        }
    }
}

void Encrypt(TStringList &inputFiles, const std::string &strOutputFilename, TElPGPKeyring &Keyring,
			 int publicKeyNum, int secretKeyNum)
{
	if (inputFiles.get_Count() == 0)
		std::cout << "Input files list is empty" << std::endl;
	else if (strOutputFilename == "" )
		std::cout << "Output file does not exist" << std::endl;
	else if (Keyring.get_PublicCount() == 0)
		std::cout << "Your keyring does not contain public keys" << std::endl;
	else if (Keyring.get_SecretCount() == 0)
		std::cout << "Your keyring does not contain secret keys" << std::endl;
	else if (publicKeyNum == -1)
		std::cout << "Please select a public key" << std::endl;
	else if (secretKeyNum == -1)
		std::cout << "Please select a secret key" << std::endl;
	else
	{
        TElPGPKeyring tempKeyring (NULL);
		tempKeyring.AddPublicKey(Keyring.get_PublicKeys(publicKeyNum));
		tempKeyring.AddSecretKey(Keyring.get_SecretKeys(secretKeyNum));

        EncryptAndSign(inputFiles, strOutputFilename, tempKeyring, "");
	}
}

void SB_CALLBACK PGP_CreateOutputStreamEvent(void * /*_ObjectData*/, TObjectHandle /*Sender*/, const char * pcFilename, int32_t szFilename, int64_t TimeStamp, TStreamHandle * Stream, int8_t * FreeOnExit)
{
	std::string filename (pcFilename, szFilename);
    if (filename != "_CONSOLE") // skip 'for eyes only' message
    {
		TFileStream fs(outputPath + filename, filemodeCreate);
		*Stream = fs.getHandle();
        *FreeOnExit = true;
		fs.updateHandle(NULL); // detach wrapped class
    }
}

void SB_CALLBACK PGP_MultipleFilesFoundEvent(void * /*_ObjectData*/, TObjectHandle /*Sender*/, const char * pcTarFilename, int32_t szTarFilename, int8_t * ExtractFiles)
{
	*ExtractFiles = true;
}

void SB_CALLBACK PGP_TemporaryStreamEvent(void * /*_ObjectData*/, TObjectHandle /*Sender*/, TStreamHandle * Stream, int8_t * FreeOnFinish)
{
	TFileStream fs(outputPath + "output.tar", filemodeCreate);
	*Stream = fs.getHandle();
    *FreeOnFinish = true;
	fs.updateHandle(NULL); // detach wrapped class
}

void DecryptAndVerify(const std::string &strInputFilename, TElPGPKeyring &Keyring, const std::string &keyPassphrase, const std::string &password)
{
	TElPGPReader pgpReader(NULL);
	if (Keyring.get_SecretCount() > 0)
		pgpReader.set_DecryptingKeys(Keyring);

	if (Keyring.get_PublicCount() > 0)
		pgpReader.set_VerifyingKeys(Keyring);

	if (keyPassphrase != "")
		pgpReader.set_KeyPassphrase(keyPassphrase);

	if (password != "")
		pgpReader.set_Passphrase(password);

	pgpReader.set_OnCreateOutputStream(&PGP_CreateOutputStreamEvent, NULL);
	pgpReader.set_OnMultipleFilesFound(&PGP_MultipleFilesFoundEvent, NULL);
	pgpReader.set_OnTemporaryStream(&PGP_TemporaryStreamEvent, NULL);
	pgpReader.set_OnKeyPassphrase(&PGP_KeyPassphraseEvent, NULL);

	TFileStream streamInput(strInputFilename, filemodeOpenReadWrite);
	try
	{
		pgpReader.DecryptAndVerify(streamInput, 0);
		std::cout << "The file was decrypted successfully" << std::endl;
	}
	catch (SBException ex)
	{
		std::cout << ex.what() << std::endl;
	}
}

void LoadKeyring(TElPGPKeyring &pgpKeyring, const std::string &publicKeyPath, const std::string &secretKeyPath, const std::string &secretPassphrase)
{
	std::cout << "Loading keyring" << std::endl;
	if ((publicKeyPath == "") || (secretKeyPath ==""))
	{
		std::cout << "Key path is empty" << std::endl;
		return;
	}

	try
	{
		pgpKeyring.Load(publicKeyPath, secretKeyPath, true);
		for (int i = 0; i < pgpKeyring.get_SecretCount(); i++)
			pgpKeyring.get_SecretKeys(i)->set_Passphrase(secretPassphrase);
	}
	catch (SBException ex)
	{
		std::cout << "Keyring loading error:" << std::endl;
		std::cout << ex.what() << std::endl;
	}
}

int main(int argc, char **argv)
{
	bool encryt = false;
	int i;
	std::string helpMessage = "Usage: consolepgp.exe [operation] input1, input2, .... [-o output] [-pk public key] [-sk secret key] []\n\n" \
		"Where possible operations are: \n" \
		"-e : encrypt input files and save to output\n" \
		"-d : decrypt input file and write files to output directory\n" \
		"Where possible flags are:\n" \
		"-pk [public key]: path to public key\n" \
		"-sk [secret key]: path to secret key\n" \
		"-sp [secret passphrase]: passphrase for secret key\n" \
		"-p password: set password for encryption/decryption\n";

	std::string publicKeyPath, secretKeyPath, secretPassphrase, password;
	int publicKeyNum = 0, secretKeyNum = 0;
	TStringList inputFiles;

	try
	{
		#error Please pick the evaluation license key from <SecureBlackbox>\LicenseKey.txt file and place it here. If the evaluation key expires, you can request an extension using the form on https://www.eldos.com/sbb/keyreq/
SetLicenseKey("<put the key here>");
	}
	catch (SBException E)
	{
		std::cout << "Failed to set license key:" << std::endl;
		std::cout << E.what() << std::endl;
		return 0;
	}

	if (argc < 2)
	{
		std::cout << helpMessage << std::endl;
		return 0;
	}

	std::cout << "EldoS OpenPGPBlackbox console demo." << std::endl;

	try
	{
		i = 1;
		while (i < argc)
		{
			std::string param;
			param.assign(argv[i]);

			if (((param == "-e") || (param == "-d")) && (i < argc - 1))
			{
				if (param == "-e")
					encryt = true;
				else
					encryt = false;

				while ((i + 1 < argc) && (argv[i + 1][0] != '-'))
				{
					i++;
					std::string name;
					name.assign(argv[i]);
					inputFiles.Add(name);
				}
			}
			else if (param == "-o")
			{
				if (++i >= argc)
					throw new SBException();

				outputPath = UnQuote(argv[i]);
			}
			else if (param == "-pk")
			{
				if (++i >= argc)
					throw new SBException();

				publicKeyPath = UnQuote(argv[i]);

				if ((i + 1 < argc) && (argv[i + 1][0] != '-'))
				{
					if (++i >= argc)
						throw new SBException();

					publicKeyNum = atoi(argv[i]);
				}
			}
			else if (param == "-sk")
			{
				if (++i >= argc)
					throw new SBException();

				secretKeyPath = UnQuote(argv[i]);

				if ((i + 1 < argc) && (argv[i + 1][0] != '-'))
				{
					if (++i >= argc)
						throw new SBException();

					secretKeyNum = atoi(argv[i]);
				}
			}
			else if (param == "-sp")
			{
				if (++i >= argc)
					throw new SBException();

				secretPassphrase = UnQuote(argv[i]);
			}
			else if (param == "-p")
			{
				if (++i >= argc)
					throw new SBException();

				password = UnQuote(argv[i]);
			}
			else if (param == "-h")
			{
				std::cout << helpMessage << std::endl;
			}
			else
				throw new SBException();

			i++;
		}
	}
	catch (...)
	{
		std::cout << "Failed to parse parameters" << std::endl;
		std::cout << helpMessage << std::endl;
		return 0;
	}

	try
	{
		TElPGPKeyring pgpKeyring(NULL);
		LoadKeyring(pgpKeyring, publicKeyPath, secretKeyPath, secretPassphrase);

		if (encryt)
		{
			EncryptAndSign(inputFiles, outputPath, pgpKeyring, password);
		}
		else
		{
			std::string inputFile;
			inputFiles.get_Strings(0, inputFile);
			DecryptAndVerify(inputFile, pgpKeyring, secretPassphrase, password);
		}
	}
	catch (SBException ex)
	{
		std::cout << "Error" << ex.what() << std::endl;
	}

	return 0;
}
