#include <cstring>
#include <iostream>
#include <fstream>
#include <time.h>
#include <stdlib.h>
#include <stdio.h>

#include "sbcustomcertstorage.h"
#include "sbmisc.h"
#include "sbpdf.h"
#include "sbpdfsecurity.h"
#include "sbutils.h"
#include "sbx509.h"

using namespace SecureBlackbox;

#define OP_VERIFY	0
#define OP_SIGN		1

// command line parameters
int paramOperation = OP_VERIFY;
TSBPDFSignatureType paramSignatureType = stDocument;
bool paramInvisibleSignature = false;
double paramOffsetX = -1;
double paramOffsetY = -1;
std::string paramAuthor;
std::string paramReason;
std::string paramCertFile;
std::string paramCertPassword;
std::string paramInputFile;
std::string paramOutputFile;

int LoadCertificate(TElX509Certificate &cert, std::string filename, std::string password)
{
	std::ifstream fs;
	fs.open(filename.c_str(), std::ios::binary | std::ios::in);
	if (fs.is_open())
	{
		TElCallbackStream cs(fs);
		int k = cert.LoadFromStreamAuto(cs, password, 0);
		fs.close();
		if (k != 0)
			std::cout << std::hex << "Failed to load certificate: 0x" << k << std::endl;

		return k;
	}
	else
	{
		std::cout << "Unable to open file: " << filename << std::endl;
		return 1;
	}
}

bool CopyFile(std::string fromFilename, std::string toFilename)
{
	std::ifstream src;
	std::ofstream dst;
	src.open(fromFilename.c_str(), std::ios::binary | std::ios::in);
	if (!src.is_open())
	{
		std::cout << "Unable to open file: " << fromFilename << std::endl;
		return false;
	}

	dst.open(toFilename.c_str(), std::ios::binary | std::ios::out);
	if (!dst.is_open())
	{
		std::cout << "Unable to open file: " << toFilename << std::endl;
		return false;
	}

	dst << src.rdbuf();

	dst.close();
	src.close();
	return true;
}

void GetTempFilename(std::string &filename)
{
	filename = "tmp.pdf";
	//filename.resize(L_tmpnam);
	//tmpnam(&filename.front());
}

void SignDocument()
{
	std::string tmpFilename;
	tmpFilename = paramInputFile;
	GetTempFilename(tmpFilename);
	if (!CopyFile(paramInputFile, tmpFilename))
	{
		std::cout << "Signing failed" << std::endl;
		return;
	}

	uint8_t Success = 0;

	// using native stream
	std::fstream fs;
	fs.open(tmpFilename.c_str(), std::ios::binary | std::ios::in | std::ios::out);
	if (fs.is_open())
	{
		TElCallbackStream cs(fs);

		TElPDFDocument Document(NULL);
		try
		{
			Document.Open(cs);
			if (Document.get_Encrypted())
			{
				std::cout << "The document is encrypted and cannot be signed" << std::endl;
				Document.Close(false);
				fs.close();
				return;
			}

			TElPDFPublicKeySecurityHandler PublicKeyHandler(NULL);
			int index = Document.AddSignature();
			TElPDFSignature *Sig = Document.get_Signatures(index);
			Sig->set_Handler(PublicKeyHandler);
			Sig->set_AuthorName(paramAuthor);
			Sig->set_Reason(paramReason);

			time_t timer;
			time(&timer);
			Sig->set_SigningTime(timer);

			Sig->set_Invisible(paramInvisibleSignature);
			Sig->set_SignatureType(paramSignatureType);

			TElX509Certificate Cert(NULL);
			if (LoadCertificate(Cert, paramCertFile, paramCertPassword))
			{
				Document.Close(false);
				fs.close();
				return;
			}

			TElMemoryCertStorage CertStorage(NULL);
			CertStorage.Add(Cert, true);

			PublicKeyHandler.set_CertStorage(CertStorage);
			PublicKeyHandler.set_SignatureType(pstPKCS7SHA1);
			PublicKeyHandler.set_CustomName("Adobe.PPKMS");

			Document.Close(true);
			Success = 1;
		}
		catch (SBException E)
		{
			std::cout << E.what() << std::endl;
			std::cout << "Stack trace: " << E.getErrorStackTrace().c_str() << std::endl;
		}

		fs.close();
	}
	else
		std::cout << "Unable to open file: " << tmpFilename << std::endl;

	if (Success)
	{
		CopyFile(tmpFilename, paramOutputFile);
		std::cout << "Signing process successfully finished" << std::endl;
	}
	else
		std::cout << "Signing failed" << std::endl;

	if (remove(tmpFilename.c_str()) != 0)
		std::cout << "Error deleting file" << std::endl;
}

void VerifyDocument()
{
	std::string tmpFilename;
	tmpFilename = paramInputFile;
	GetTempFilename(tmpFilename);
	if (!CopyFile(paramInputFile, tmpFilename))
	{
		std::cout << "Verification failed" << std::endl;
		return;
	}

	//uint8_t DocumentChanged = 0;
	try
	{
		// using library stream
		TFileStream fs(tmpFilename, filemodeOpenReadWrite);
		try
		{
			TElPDFDocument Document(NULL);
			Document.Open(fs);

			if (!Document.get_Encrypted() && !Document.get_Signed())
			{
				std::cout << "Document is neither encrypted nor signed" << std::endl;
				Document.Close(false);
				return;
			}

			if (Document.get_Signed())
			{
				for (int i = 0; i < Document.get_SignatureCount(); i++)
				{
					TElPDFSignature *Sig = Document.get_Signatures(i);
					std::string s;
					Sig->get_SignatureName(s);
					std::cout << "Signature name: " << s << std::endl;

					std::cout << "Author name: ";
					Sig->get_AuthorName(s);
					if (s.empty())
						std::cout << "<not specified>" << std::endl;
					else
						std::cout << s << std::endl;

					std::cout << "Reason: ";
					Sig->get_Reason(s);
					if (s.empty())
						std::cout << "<not specified>" << std::endl;
					else
						std::cout << s << std::endl;

					time_t timer = (time_t)Sig->get_SigningTime();
					char* dt = ctime(&timer);
					std::cout << "Signing time: " << dt << std::endl;
					if (Sig->Validate(true))
					{
						// todo: validte signing certificate
						std::cout << "The selected signature is VALID" << std::endl;
					}
					else
						std::cout << "The selected signature is NOT VALID" << std::endl;
				}
			}
		}
		catch (SBException E)
		{
			std::cout << E.what() << std::endl;
			std::cout << "Stack trace: " << E.getErrorStackTrace().c_str() << std::endl;
		}
	}
	catch (SBException E)
	{
		std::cout << E.what() << std::endl;
		std::cout << "Stack trace: " << E.getErrorStackTrace().c_str() << std::endl;
	}

	if (remove(tmpFilename.c_str()) != 0)
		std::cout << "Error deleting file" << std::endl;
}

int main(int argc, char **argv) {

	try
	{
		#error Please pick the evaluation license key from <SecureBlackbox>\LicenseKey.txt file and place it here. If the evaluation key expires, you can request an extension using the form on https://www.eldos.com/sbb/keyreq/
SetLicenseKey("<put the key here>");
	}
	catch (SBException E)
	{
		std::cout << "Failed to set license key:" << std::endl;
		std::cout << E.what() << std::endl;
		return 0;
	}

	if (argc < 2)
	{
		std::cout << "Usage:" << std::endl;
		std::cout << "ConsolePDF [operation] [signature options] [key options] input.pdf [-out output.pdf]" << std::endl <<
			"Operation: " << std::endl <<
			" -sign : sign pdf document" << std::endl <<
			" -verify : verify pdf document" << std::endl <<
			"Signature options: " << std::endl <<
			" -invis | -vis | -mdp : signature type (invisible, visible, mdp)" << std::endl <<
			//" -offsetx value | -offsety value : signature widget offset" << std::endl <<
			" -author value : author name" << std::endl <<
			" -reason value : signing reason" << std::endl <<
			"Key options:" << std::endl <<
			" -cert filename : certificate filename" << std::endl <<
			" -certpass password : certificate password" << std::endl <<
			std::endl;

		return 0;
	}

	int i = 1;
	while (i < argc - 1)
	{
		char *arg = argv[i];
		if (!strcmp(arg, "-sign"))
			paramOperation = OP_SIGN;
		else if (!strcmp(arg, "-verify"))
			paramOperation = OP_VERIFY;
		else if (!strcmp(arg, "-invis"))
			paramInvisibleSignature = true;
		else if (!strcmp(arg, "-vis"))
			paramInvisibleSignature = false;
		else if (!strcmp(arg, "-mdp"))
			paramSignatureType = stMDP;
		else if (!strcmp(arg, "-cert"))
			paramCertFile.assign(argv[++i]);
		else if (!strcmp(arg, "-certpass"))
			paramCertPassword.assign(argv[++i]);
		else if (!strcmp(arg, "-offsetx"))
			paramOffsetX = atof(argv[++i]);
		else if (!strcmp(arg, "-offsety"))
			paramOffsetY = atof(argv[++i]);
		else if (!strcmp(arg, "-author"))
			paramAuthor.assign(argv[++i]);
		else if (!strcmp(arg, "-reason"))
			paramReason.assign(argv[++i]);
		else if (!strcmp(arg, "-out"))
			paramOutputFile.assign(argv[++i]);
		else
			paramInputFile.assign(argv[i]);

		i++;
	}

	if (i < argc)
		paramInputFile.assign(argv[i]);

	if (paramInputFile.empty())
	{
		std::cout << "Input filename is not specified" << std::endl;
		return 0;
	}

	if (paramOperation != OP_VERIFY)
	{
		if (paramOutputFile.empty())
		{
			std::cout << "Output filename is not specified" << std::endl;
			return 0;
		}

		if (paramCertFile.empty())
		{
			std::cout << "Certificate filename is not specified" << std::endl;
			return 0;
		}
	}

	switch (paramOperation)
	{
	case OP_SIGN:
		SignDocument();
		break;
	default:
		VerifyDocument();
		break;
	}

	return 0;
}
